"""The kernels for XGrammar. There are 5 implementations:

- CPU: used for CPU tensors
- CUDA: not used in the current implementation
- Triton: used for CUDA GPU tensors
- MLX: used for MLX tensors
- Torch Compile: used for torch tensors on other devices
"""
