/* Fix any overlay issues in tlcpack theme that might prevent text selection */
.wy-grid-for-nav {
    position: relative !important;
}

/* Make header always fixed/floating at the top */
.header {
    position: fixed !important;
    top: 0 !important;
    left: 0 !important;
    width: 100% !important;
    z-index: 999 !important;
    box-shadow: 0 2px 8px rgba(0, 0, 0, 0.15) !important;
}

.wy-nav-side .wy-side-scroll {
    padding-top: 25px !important;
}

.wy-nav-side.fixed {
    top: 0 !important;
    padding-top: 54px !important;
    padding-bottom: 0 !important;
}

/* Adjust content wrapper for fixed header */
.wy-nav-content-wrap {
    padding-top: 75px !important;
}
